#include <fstream>
#include "KiteLexer.hpp"
#include "KiteParser.hpp"
#include "KiteWalker.hpp"
#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteCode.h"
#include "KiteIdentifier.h"
#include "KiteBoolean.h"
#include "KiteString.h"
#include "KiteList.h"
#include "KiteKeyValPair.h"
#include "KiteTree.h"
#include "KiteFile.h"
#include "KiteDynamicObject.h"

using namespace antlr;

namespace KiteVM {
	KiteReference<KiteObject> LoadInstruction::eval(VM &vm) {
		KiteReference<KiteObject> name(vm.stack.top());
		vm.stack.pop();
        
        fstream file;
        file.open(dynamic_cast<KiteString&>(*name).getString().c_str(), ios::in);
        
        KiteLexer lexer(file);
        KiteParser parser(lexer);
        ASTFactory my_factory;
        parser.initializeASTFactory( my_factory );
        parser.setASTFactory( &my_factory );
        parser.expr();
        RefAST t = parser.getAST();
        
        if (t.get()) {
            KiteWalker walker;
            Instruction *i = walker.expr(t);
            vm.run(i);
            i->deallocate();
            return vm.stack.top();
        }
        
        return KiteReference<KiteObject>(new KiteTree("Parse error while compiling code"));
	}

	KiteReference<KiteObject> VarInstruction::eval(VM &vm) {
		KiteReference<KiteIdentifier> lhs(vm.stack.top().referenceIdentifier());
		vm.stack.pop();
		KiteReference<KiteObject> rhs(vm.stack.top());
		vm.stack.pop();

		vm.syms.addVar(lhs, rhs);
		return rhs;
	}

	KiteReference<KiteObject> ClassInstruction::eval(VM &vm) {
		KiteReference<KiteObject> name(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> definition(vm.stack.top());
		vm.stack.pop();

		if (typeid(*definition) == typeid(KiteDynamicObject)) {
			// Simple assignment
			vm.syms.addVar(name.referenceIdentifier(), definition);
            dynamic_cast<KiteDynamicObject&>(*definition).type = (*name.referenceIdentifier()).getString();
            return definition;
		} else if (typeid(*definition) == typeid(KiteCode)) {
			KiteCode &c = dynamic_cast<KiteCode&>(*definition);

			// Create class
			KiteReference<KiteObject> ret(new KiteDynamicObject(vm, c.getCode()));
			if (typeid(*ret) != typeid(KiteTree)) {
				vm.syms.addVar(name.referenceIdentifier(), ret);
			}
            
            dynamic_cast<KiteDynamicObject&>(*ret).type = (*name.referenceIdentifier()).getString();
			return ret;
		}
        
        return KiteReference<KiteObject>(new KiteTree("Class requires a code object"));
	}

	KiteReference<KiteObject> FunctionCallInstruction::eval(VM &vm) {
		KiteReference<KiteObject> lhs(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> name(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> args(vm.stack.top());
		vm.stack.pop();

		KiteReference<KiteObject> newlhs, newrhs;
		if (typeid(*lhs) == typeid(KiteIdentifier) &&
            dynamic_cast<KiteIdentifier&>(*lhs).getString() != "tree" &&
            dynamic_cast<KiteIdentifier&>(*lhs).getString() != "file") {
			// Gotta find the lhs in the symbol table first
			newlhs = vm.syms.get(lhs.referenceIdentifier());
		} else {
			newlhs = lhs;
		}

		if (typeid(*args) == typeid(KiteIdentifier)) {
			// Gotta find the rhs in the symbol table first
			newrhs = vm.syms.get(args.referenceIdentifier());
		} else {
			newrhs = args;
		}

		// If a class object (and new was called), create new instance
		if ((*name.referenceIdentifier()).getString() == "new") {
            if (typeid(*lhs) == typeid(KiteIdentifier) &&
                dynamic_cast<KiteIdentifier&>(*lhs).getString() == "tree") {
                string msg = dynamic_cast<KiteString&>(*coerceString(vm, newrhs)).getString();
                KiteReference<KiteObject> ret(new KiteTree(msg));
                return ret;
            } else if (typeid(*lhs) == typeid(KiteIdentifier) &&
                dynamic_cast<KiteIdentifier&>(*lhs).getString() == "file") {
                KiteReference<KiteObject> ret(new KiteFile(vm, args));
                return ret;
            }
			KiteReference<KiteObject> ret((*newlhs).create(vm, newlhs, args));
			return ret;
        } else if ((*name.referenceIdentifier()).getString() == "op+"
                 && typeid(*newlhs) == typeid(KiteDynamicObject)
                 && !dynamic_cast<KiteDynamicObject&>(*newlhs).isInstance()
                 && typeid(*newrhs) == typeid(KiteCode)) {
            // New class with LHS as parent class
            KiteReference<KiteObject> ret(new KiteDynamicObject(vm, newlhs, dynamic_cast<KiteCode&>(*newrhs).getCode()));
            return ret;
		} else {
			vm.syms.addContext();
            //if (typeid(*lhs) == typeid(KiteIdentifier) && !(*lhs.referenceIdentifier() == KiteIdentifier("this"))) {
            //if (newlhs.getObj() != vm.syms.get(KiteReference<KiteIdentifier>(new KiteIdentifier("this"))).getObj())
                vm.syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("this")), newlhs);
            //}
            KiteReference<KiteObject> fun((*newlhs).dereference(vm, name.referenceIdentifier()));
            KiteReference<KiteObject> ret((*fun).call(vm, newrhs));
			vm.syms.removeContext();
			return ret;
		}
	}

	KiteReference<KiteObject> StringInstruction::eval(VM &vm) {
		return KiteReference<KiteObject>(new KiteString(constant));
	}

	KiteReference<KiteObject> IdentifierInstruction::eval(VM &vm) {
		return KiteReference<KiteObject>(new KiteIdentifier(constant));
	}

	KiteReference<KiteObject> BooleanInstruction::eval(VM &vm) {
		return KiteReference<KiteObject>(new KiteBoolean(constant));
	}

	KiteReference<KiteObject> CodeInstruction::eval(VM &vm) {
		return KiteReference<KiteObject>(new KiteCode(constant));
	}

	KiteReference<KiteObject> NumberInstruction::eval(VM &vm) {
		return KiteReference<KiteObject>(new KiteNumber(constant));
	}

	KiteReference<KiteObject> AssociateArrayInstruction::eval(VM &vm) {
		KiteReference<KiteObject> key(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> val(vm.stack.top());
		vm.stack.pop();

		return KiteReference<KiteObject>(new KiteKeyValPair(key, val));
	}

	KiteReference<KiteObject> AssignInstruction::eval(VM &vm) {
		KiteReference<KiteObject> lhs(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> rhs(vm.stack.top());
		vm.stack.pop();

		if (typeid(*lhs) == typeid(KiteIdentifier)) {
			return (vm.syms.replace(lhs.referenceIdentifier(), rhs));
		} else {
			lhs.replace(rhs);
			return lhs;
		}
	}

	KiteReference<KiteObject> ArrayDereferenceInstruction::eval(VM &vm) {
		KiteReference<KiteObject> name(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> idx(vm.stack.top());
		vm.stack.pop();

		KiteReference<KiteObject> newlhs, newrhs;
		if (typeid(*name) == typeid(KiteIdentifier)) {
			// Gotta find the lhs in the symbol table first
			newlhs = vm.syms.get(name.referenceIdentifier());
		} else {
			newlhs = name;
		}

		if (typeid(*idx) == typeid(KiteIdentifier)) {
			// Gotta find the lhs in the symbol table first
			newrhs = vm.syms.get(idx.referenceIdentifier());
		} else {
			newrhs = idx;
		}

		vm.syms.addContext();
		vm.syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("this")), newlhs);
		KiteReference<KiteObject> fun((*newlhs).dereference(vm, KiteReference<KiteIdentifier>(new KiteIdentifier("opderef"))));
		KiteReference<KiteObject> ret((*fun).call(vm, newrhs));
		vm.syms.removeContext();
		return ret;
	}

	KiteReference<KiteObject> EmptyArrayInstruction::eval(VM &vm) {
		return KiteReference<KiteObject>(new KiteList());
	}

	KiteReference<KiteObject> MergeArrayInstruction::eval(VM &vm) {
		KiteReference<KiteObject> lhs(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> rhs(vm.stack.top());
		vm.stack.pop();

		if (typeid(*lhs) == typeid(KiteList) && typeid(*rhs) != typeid(KiteList)) {
			KiteList list1(dynamic_cast<KiteList&>(*lhs));
			KiteList list2(rhs);
			return KiteReference<KiteObject>(list1.merge(list2));
		} else if (typeid(*rhs) == typeid(KiteList) && typeid(*lhs) != typeid(KiteList)) {
			KiteList list2(dynamic_cast<KiteList&>(*rhs));
			KiteList list1(lhs);
			return KiteReference<KiteObject>(list1.merge(list2));
		} else if (typeid(*rhs) != typeid(KiteList) && typeid(*lhs) != typeid(KiteList)) {
			KiteList list2(rhs);
			KiteList list1(lhs);
			return KiteReference<KiteObject>(list1.merge(list2));
		} else {
			KiteList list1(dynamic_cast<KiteList&>(*lhs));
			KiteList list2(dynamic_cast<KiteList&>(*rhs));
			return KiteReference<KiteObject>(list1.merge(list2));
		}
	}

	KiteReference<KiteObject> DereferenceInstruction::eval(VM &vm) {
		KiteReference<KiteObject> lhs(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> rhs(vm.stack.top());
		vm.stack.pop();

		if (typeid(*lhs) == typeid(KiteIdentifier)) {
			// Gotta find the lhs in the symbol table first
			KiteReference<KiteObject> newlhs(vm.syms.get(lhs.referenceIdentifier()));
			return (*newlhs).dereference(vm, rhs.referenceIdentifier());
		}

		return (*lhs).dereference(vm, rhs.referenceIdentifier());
	}

	KiteReference<KiteObject> VM::run(KiteVM::Instruction *instr) {
		while(instr) {
			stack.push(instr->eval(*this));

			if (typeid(*stack.top()) == typeid(KiteTree)) {
				break;
			}

			instr = instr->next;
		}

		return stack.top();
	}

	VM::~VM() {
		// nothing required
	}

	VM::VM() {
        KiteObject::getClassinfo();
		syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("stdin")), KiteReference<KiteObject>(new KiteFile(&std::cin)));
		syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("stdout")), KiteReference<KiteObject>(new KiteFile(&std::cout)));
		syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("stderr")), KiteReference<KiteObject>(new KiteFile(&std::cerr)));
	}
};

KiteReference<KiteObject> coerceNumber(KiteVM::VM &vm, KiteReference<KiteObject> obj) {
	vm.syms.addContext();
	vm.syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("this")), obj);
	KiteReference<KiteObject> fun((*obj).dereference(vm, KiteReference<KiteIdentifier>(new KiteIdentifier("number"))));
	KiteReference<KiteObject> ret((*fun).call(vm, KiteReference<KiteObject>(NULL)));
	vm.syms.removeContext();
	return ret;
}

KiteReference<KiteObject> coerceString(KiteVM::VM &vm, KiteReference<KiteObject> obj) {
	vm.syms.addContext();
	vm.syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("this")), obj);
	KiteReference<KiteObject> fun((*obj).dereference(vm, KiteReference<KiteIdentifier>(new KiteIdentifier("string"))));
	KiteReference<KiteObject> ret((*fun).call(vm, KiteReference<KiteObject>(NULL)));
	vm.syms.removeContext();
	return ret;
}

bool valuesEqual(KiteVM::VM &vm, KiteReference<KiteObject> x, KiteReference<KiteObject> j) {
	vm.syms.addContext();
	vm.syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("this")), x);
	KiteReference<KiteObject> fun((*x).dereference(vm, KiteReference<KiteIdentifier>(new KiteIdentifier("op=="))));
	KiteReference<KiteObject> ret((*fun).call(vm, j));
	vm.syms.removeContext();

	if (typeid(*ret) == typeid(KiteTree)) {
		// Problem encountered; not equal.
		return false;
	} else if (typeid(*ret) == typeid(KiteBoolean)) {
		return dynamic_cast<KiteBoolean&>(*ret).boolVal();
	} else {
		// Unknown error
		return false;
	}
}