#ifndef KITEVM_H
#define KITEVM_H

#include <iostream>
#include <string>
#include <stack>
#include "KiteReference.h"
#include "SymbolTable.h"

using namespace std;

class KiteObject;

extern string processEscapes(string str);

namespace KiteVM {
	class Instruction;
    
    class FuncCall {
    public:
        FuncCall() { }
        
        FuncCall(string name, string file, int line)
        : name(name), file(file), line(line) { }
        
        FuncCall(const FuncCall &obj)
        : name(obj.name), file(obj.file), line(obj.line) { }
    
        string name;
        string file;
        int line;
    };
    
	class VM {
	public:
		VM();
		virtual ~VM();

		KiteReference<KiteObject> run(Instruction *instr);
	//private:
		stack<KiteReference<KiteObject> > stack; // This _shouldn't_ be public.
        vector<FuncCall> funcCalls;
		SymbolTable syms;
	};

	class Instruction {
	public:
		Instruction();
		virtual ~Instruction();

		void addToEnd(Instruction *p);
		virtual KiteReference<KiteObject> eval(VM &vm) = 0;
		virtual std::string toString() const = 0;

		void deallocate();
		void outTree();

		friend class VM;
	protected:
		Instruction *next;
	};

	class SingleArgumentInstruction : public Instruction {
	public:
		SingleArgumentInstruction(Instruction *arg1) {
			arg1->addToEnd(this);
		}

		virtual KiteReference<KiteObject> eval(VM &vm) = 0;
	};

	class LoadInstruction : public SingleArgumentInstruction {
	public:
		LoadInstruction(Instruction *arg1) : SingleArgumentInstruction(arg1) { }
		virtual std::string toString() const { return "LOAD"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class DefinitionInstruction : public Instruction {
	public:
		DefinitionInstruction(Instruction *lhs, Instruction *rhs) {
			rhs->addToEnd(lhs);
			lhs->addToEnd(this);
		}

		virtual KiteReference<KiteObject> eval(VM &vm) = 0;
	};

	class VarInstruction : public DefinitionInstruction {
	public:
		VarInstruction(Instruction *lhs, Instruction *rhs) : DefinitionInstruction(lhs, rhs) { }
		virtual std::string toString() const { return "DEFINEVAR"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class ClassInstruction : public DefinitionInstruction {
	public:
		ClassInstruction(Instruction *lhs, Instruction *rhs) : DefinitionInstruction(lhs, rhs) { }
		virtual std::string toString() const { return "DEFINECLASS"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class FunctionCallInstruction : public Instruction {
	public:
		FunctionCallInstruction(Instruction *object, Instruction *name, Instruction *arguments)  {
			if (arguments) arguments->addToEnd(name);
			name->addToEnd(object);
			object->addToEnd(this);
		}

		virtual std::string toString() const { return "CALL"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	template<class T>
	class ConstantInstruction : public Instruction {
	public:
		ConstantInstruction(T val)
			: constant(val) { }

		virtual KiteReference<KiteObject> eval(VM &vm) = 0;
	protected:
		T constant;
	};

	class StringInstruction : public ConstantInstruction<std::string> {
	public:
		StringInstruction(const char *string) : ConstantInstruction<std::string>(processEscapes(std::string(string))) { }
		virtual std::string toString() const { return "STRING " + constant; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class IdentifierInstruction : public ConstantInstruction<std::string> {
	public:
		IdentifierInstruction(const char *string) : ConstantInstruction<std::string>(std::string(string)) { }
		virtual std::string toString() const { return "IDENTIFIER " + constant; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class NumberInstruction : public ConstantInstruction<std::string> {
	public:
		NumberInstruction(const char *string) : ConstantInstruction<std::string>(std::string(string)) { }
		virtual std::string toString() const { return "NUMBER " + constant; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class BooleanInstruction : public ConstantInstruction<std::string> {
	public:
		BooleanInstruction(bool val) : ConstantInstruction<std::string>(val ? std::string("true") : std::string("false")) { }
		virtual std::string toString() const { return "BOOLEAN " + constant; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class CodeInstruction : public ConstantInstruction<Instruction*> {
	public:
		CodeInstruction(Instruction *code) : ConstantInstruction<Instruction*>(code) { }
		virtual ~CodeInstruction() { constant->deallocate(); }
		virtual std::string toString() const { 
			std::string ret = "CODE {\n";
			/*Instruction *cur = constant;
			while(cur) {
				ret += cur->toString() + "\n";
				cur = cur->next;
			}*/
			return ret + "\n}";
		}

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class AssociateArrayInstruction : public DefinitionInstruction {
	public:
		AssociateArrayInstruction(Instruction *lhs, Instruction *rhs) : DefinitionInstruction(lhs, rhs) { }
		virtual std::string toString() const { return "ASSOCIATE"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class AssignInstruction : public DefinitionInstruction {
	public:
		AssignInstruction(Instruction *lhs, Instruction *rhs) : DefinitionInstruction(lhs, rhs) { }
		virtual std::string toString() const { return "ASSIGN"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class ArrayDereferenceInstruction : public DefinitionInstruction {
	public:
		ArrayDereferenceInstruction(Instruction *lhs, Instruction *rhs) : DefinitionInstruction(lhs, rhs) { }
		virtual std::string toString() const { return "ARRAYELEMENT"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class EmptyArrayInstruction : public Instruction {
	public:
		virtual std::string toString() const { return "EMPTYARRAY"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class MergeArrayInstruction : public DefinitionInstruction {
	public:
		MergeArrayInstruction(Instruction *lhs, Instruction *rhs) : DefinitionInstruction(lhs, rhs) { }
		virtual std::string toString() const { return "MERGEARRAY"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};

	class DereferenceInstruction : public DefinitionInstruction {
	public:
		DereferenceInstruction(Instruction *lhs, Instruction *rhs) : DefinitionInstruction(lhs, rhs) { }
		virtual std::string toString() const { return "DEREFERENCE"; }

		virtual KiteReference<KiteObject> eval(VM &vm);
	};
};

#define THIS_OBJECT() (vm.syms.get(KiteReference<KiteIdentifier>(new KiteIdentifier("this"))))

KiteReference<KiteObject> coerceNumber(KiteVM::VM &vm, KiteReference<KiteObject> obj);
KiteReference<KiteObject> coerceString(KiteVM::VM &vm, KiteReference<KiteObject> obj);
bool valuesEqual(KiteVM::VM &vm, KiteReference<KiteObject> x, KiteReference<KiteObject> j);

#endif // KITEVM_H
