header "post_include_hpp" {
    #include <iostream>
    #include "KiteVM.h"
    #include "KiteIdentifier.h"
    #include "KiteObject.h"
    using namespace std;
    using namespace KiteVM;
}

options {    
    language="Cpp";
}

class KiteParser extends Parser;

options {
    k = 2; // Not an LL(1) grammar due to array syntax
    buildAST = true;
}

expr: (singleexpr SEMICOLON^)*
     ;

singleexpr:
     loadexpr
     | varexpr
     | classexpr
     | assignmentexpr
     ;

loadexpr:
     "load"^ STRING
     ;

varexpr:
     "var"^ IDENTIFIER (EQUALS! computedexpr)?
     ;

classexpr:
     "class"^ IDENTIFIER EQUALS! computedexpr
     ;

assignmentexpr:
     computedexpr (EQUALS^ computedexpr)*
     ;

computedexpr:
     orexpr ("or"^ orexpr)*
     ;

orexpr:
     andexpr ("and"^ andexpr)*
     ;

andexpr:
     equalexpr ((EQUALITY^|NOTEQUALS^) equalexpr)*
     ;

equalexpr:
     compareexpr ((LESSTHAN^|LE^|GT^|GE^) compareexpr)*
     ;

compareexpr:
     bitwiseexpr ((LEFTSHIFT^|RIGHTSHIFT^) bitwiseexpr)*
     ;

bitwiseexpr:
     multiplyexpr ((PLUS^|MINUS^) multiplyexpr)*
     ;

multiplyexpr:
     unaryexpr ((STAR^|DIVIDE^|"mod"^) unaryexpr)*
     ;

unaryexpr:
     (MINUS^|"not"^)? dereferenceexpr
     ;

dereferenceexpr:
     atom (LBRACKET^ (NUMBER|STRING|IDENTIFIER) RBRACKET | DEREFERENCE^ IDENTIFIER | FUNCCALL^ IDENTIFIER funcparams)*
     ;

funcparams:
     | LPAREN! arglist RPAREN!
     ;

arglist: computedexpr (COMMA^ computedexpr)*
     ;

atom:
     NUMBER
     | STRING
     | IDENTIFIER
     |! (IDENTIFIER LPAREN arglist RPAREN) => i:IDENTIFIER LPAREN^ j:arglist RPAREN { #atom = #([FUNCCALL, "|"], i, [IDENTIFIER, "eval"], j); }
     |! k:IDENTIFIER LPAREN^ RPAREN { #atom = #([FUNCCALL, "|"], k, [IDENTIFIER, "eval"]); }
     | "true"
     | "false"
     | LBRACKET^ arraydef RBRACKET!
     | LCURLY^ expr RCURLY
     | LPAREN! assignmentexpr RPAREN!
     ;

arrayelement: ((NUMBER|STRING|IDENTIFIER) COLON^)? computedexpr 
     ;

arraydef: (arrayelement (COMMA^ arrayelement)*)?
     ;

class KiteWalker extends TreeParser;

options {
    importVocab = KiteParser;
}

expr returns [Instruction *obj = NULL]
{ Instruction *a, *b, *c; }
    : #("load" a=expr) { obj = a; new LoadInstruction(a); }
    | #("var" a=expr b=expr) { obj = b; new VarInstruction(a, b); }
    | #("class" a=expr b=expr) { obj = b; new ClassInstruction(a, b); }
    | #(EQUALS a=expr b=expr) { obj = b; new AssignInstruction(a, b); }
    | #("or" a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("opor"), b); }
    | #("and" a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("opand"), b); }
    | #(EQUALITY a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op=="), b); }
    | #(NOTEQUALS a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op!="), b); }
    | #(LESSTHAN a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op<"), b); }
    | #(LE a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op<="), b); }
    | #(GT a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op>"), b); }
    | #(GE a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op>="), b); }
    | #(LEFTSHIFT a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op<<"), b); }
    | #(RIGHTSHIFT a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op>>"), b); }
    | #(PLUS a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op+"), b); }
    | (#(MINUS expr expr)) => #(MINUS a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op-"), b); }
    | #(MINUS a=expr) { obj = b; new FunctionCallInstruction(new NumberInstruction("0"), new IdentifierInstruction("op-"), a); }
    | #(STAR a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op*"), b); }
    | #(DIVIDE a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("op/"), b); }
    | #("mod" a=expr b=expr) { obj = b; new FunctionCallInstruction(a, new IdentifierInstruction("opmod"), b); }
    | #("not" a=expr) { obj = a; new FunctionCallInstruction(a, new IdentifierInstruction("opnot"), new EmptyArrayInstruction()); }
    | (#(LBRACKET expr expr RBRACKET)) => #(LBRACKET a=expr b=expr RBRACKET) { obj = b; new ArrayDereferenceInstruction(a, b); }
    | (#(LBRACKET expr)) => #(LBRACKET a=expr) { obj = a; new MergeArrayInstruction(new EmptyArrayInstruction(), a); }
    | LBRACKET { obj = new EmptyArrayInstruction(); }
    | #(COLON a=expr b=expr) { obj = b; new AssociateArrayInstruction(a, b); }
    | #(DEREFERENCE a=expr b=expr) { obj = b; new DereferenceInstruction(a, b); }
    | (#(FUNCCALL expr expr expr)) => #(FUNCCALL a=expr b=expr c=expr) { obj = c; new FunctionCallInstruction(a, b, c); }
    | #(FUNCCALL a=expr b=expr) { obj = new EmptyArrayInstruction(); new FunctionCallInstruction(a, b, obj); }
    | i:NUMBER { obj = new NumberInstruction(i->getText().c_str()); }
    | j:STRING { obj = new StringInstruction(j->getText().c_str()); }
    | k:IDENTIFIER { obj = new IdentifierInstruction(k->getText().c_str()); }
    | "true" { obj = new BooleanInstruction(true); }
    | "false" { obj = new BooleanInstruction(false); }
    | #(LCURLY a=expr RCURLY) { obj = new CodeInstruction(a); }
    | #(COMMA a=expr b=expr) { obj = b; new MergeArrayInstruction(a, b); }
    | (#(SEMICOLON expr expr)) => #(SEMICOLON a=expr b=expr) { obj = a; a->addToEnd(b); }
    | #(SEMICOLON a=expr) { obj = a; }
    ;

class KiteLexer extends Lexer;

options {
    k=2; // length of longest reserved word
    charVocabulary='\u0000'..'\u007F'; // allow ascii
}

{
public:
	void uponEOF() {
		// Nothing needed
	}
}

NUMBER: ('0'..'9')+ ('.' ('0'..'9')+ (('E'|'e') ('-')? ('0'..'9')+)?)*;
IDENTIFIER: ('A'..'Z'|'a'..'z'|'_') (~(' ' | '\r' | '\n' | '\t' | ';' | '(' | '|' | '.' | ')' | '['))*;
STRING: '"' (~('"'|'\\')|'\\' ('\\'|'r'|'n'|'"'|'t'|'x' '0'..'9'|'0'..'9'))* '"';
PLUS  : '+' ;
MINUS : '-' ; 
STAR  : '*' ;
DIVIDE: '/' ;
SEMICOLON: ';' ;
DEREFERENCE: '.';
FUNCCALL: '|';
LEFTSHIFT: "<<";
RIGHTSHIFT: ">>";
LESSTHAN: '<';
LE: "<=";
GT: '>';
GE: ">=";
EQUALITY: "==";
NOTEQUALS: "!=";
LCURLY: '{';
RCURLY: '}';
EQUALS: '=';
LPAREN: '(';
RPAREN: ')';
COMMA: ',';
COLON: ':';
LBRACKET: '[';
RBRACKET: ']';
COMMENT : '#' (~'\n')* '\n' {$setType(antlr::Token::SKIP);};
WS    : ( ' '
        | '\r' '\n'
        | '\n'
        | '\t'
        )
        {$setType(antlr::Token::SKIP);}
        ;
